﻿using Microsoft.Xna.Framework;
using xTile.Dimensions;
using StardewValley;
using StardewValley.Projectiles;
using StardewValley.TerrainFeatures;

namespace BossCreatures
{
	public class BossProjectile : BasicProjectile
	{
		private readonly bool pullPlayerIn;
		private Vector2 startingPosition;
		public bool ignoreTreeCollisions = false;

		public BossProjectile()
		{
		}

		public BossProjectile(int damageToFarmer, int parentSheetIndex, int bouncesTillDestruct, int tailLength, float rotationVelocity, float xVelocity, float yVelocity, Vector2 startingPosition, string collisionSound, string bounceSound, string firingSound, bool explode, bool damagesMonsters = false, GameLocation location = null, Character firer = null, onCollisionBehavior collisionBehavior = null, string shotItemId = null, string debuff = null, bool pullIn = false) : base(damageToFarmer, parentSheetIndex, bouncesTillDestruct, tailLength, rotationVelocity, xVelocity, yVelocity, startingPosition, collisionSound, bounceSound, firingSound, explode, damagesMonsters, location, firer, collisionBehavior, shotItemId)
		{
			this.debuff.Value = debuff;
			this.startingPosition = startingPosition;

			pullPlayerIn = pullIn;
		}

		public override bool isColliding(GameLocation location, out Character target, out TerrainFeature terrainFeature)
		{
			target = null;
			terrainFeature = null;
            Microsoft.Xna.Framework.Rectangle boundingBox = getBoundingBox();
			if (!ignoreCharacterCollisions.Value)
			{
				if (damagesMonsters.Value)
				{
					Character character = location.doesPositionCollideWithCharacter(boundingBox);
					if (character != null)
					{
						if (character is NPC && (character as NPC).IsInvisible)
						{
							return false;
						}

						target = character;
						return true;
					}
				}
				else if (Game1.player.currentLocation == location && Game1.player.GetBoundingBox().Intersects(boundingBox))
				{
					target = Game1.player;
					return true;
				}
			}

			foreach (Vector2 item in Utility.getListOfTileLocationsForBordersOfNonTileRectangle(boundingBox))
			{
				if (location.terrainFeatures.TryGetValue(item, out TerrainFeature value) && !value.isPassable() && !ignoreTreeCollisions)
				{
					terrainFeature = value;
					return true;
				}
			}

			if (!location.isTileOnMap(position.Value / 64f) || (!ignoreLocationCollision.Value && location.isCollidingPosition(boundingBox, Game1.viewport, isFarmer: false, 0, glider: true, theOneWhoFiredMe.Get(location), pathfinding: false, projectile: true)))
			{
				return true;
			}

			return false;
		}

		public override void behaviorOnCollisionWithPlayer(GameLocation location, Farmer player)
		{
			base.behaviorOnCollisionWithPlayer(location, player);
			if (debuff.Value != null)
			{
				player.applyBuff(new BossDebuff(debuff.Value));
			}
			if (pullPlayerIn)
			{
				Vector2 newPos = player.Position + (startingPosition - player.Position) * 0.02f;
				if(location.isTilePassable(new Location((int)(newPos.X/64), (int)(newPos.Y/64)), Game1.viewport))
				{
					player.Position = newPos;
				}
			}
		}
	}
}
