﻿using System;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Netcode;
using StardewValley;
using StardewValley.Monsters;
using StardewValley.Network;
using StardewValley.Projectiles;
using StardewValley.Extensions;
using xTile.Tiles;

namespace BossCreatures
{
	public class PlantGirlBoss : SquidKid
	{
		private readonly float difficulty;
		private readonly int width;
		private readonly int height;
		private float lastIceBall;
		private int maxIceBallShield;
		private int currentIceBallShield = 0;
		private int shieldTimer = 0;

		public int timeUntilNextAttack;
		public readonly NetBool firing = new(false);
		public NetInt attackState = new();
		public int nextFireTime;
		public int totalFireTime;

		public PlantGirlBoss()
		{
		}

		public PlantGirlBoss(Vector2 spawnPos, float _difficulty) : base(spawnPos)
		{
			isGlider.Value = false;
			width = ModEntry.Config.PlantGirlBossWidth;
			height = ModEntry.Config.PlantGirlBossHeight;
			Sprite.SpriteWidth = width;
			Sprite.SpriteHeight = height;
			Sprite.LoadTexture(ModEntry.GetBossTexture(GetType()));
			Sprite.UpdateSourceRect();
			Scale = ModEntry.Config.PlantGirlBossScale;
			Sprite.framesPerAnimation = 4;
			ignoreMovementAnimations = true;
			speed = 1;
			defaultAnimationInterval.Value = 150;
			experienceGained.Value = 300;

			difficulty = _difficulty;
			Health = (int)Math.Round(Health * 1500 * difficulty);
			MaxHealth = Health;
			DamageToFarmer = (int)Math.Round(damageToFarmer.Value * 2 * difficulty);
			farmerPassesThrough = true;
			moveTowardPlayerThreshold.Value = 40;
		}

		private int projectileCount = 0;
		private float heldObjectBobTimer;

		int lastTotalGameTime = 0;
		int lastDirection = 0;
		protected override void updateAnimation(GameTime time)
		{
			int currentFrame = Sprite.currentFrame;
			lastTotalGameTime += time.ElapsedGameTime.Milliseconds;
			if (lastTotalGameTime > 500)
			{
				lastTotalGameTime = 0; lastDirection = facingDirection.Get();
			}
			switch (lastDirection)
			{
				case 0: Sprite.AnimateUp(time, 1); break;
				case 1: Sprite.AnimateRight(time, 1); break;
				case 3: Sprite.AnimateLeft(time, 1); break;
				default: Sprite.AnimateDown(time, 1); break;
			}

			if (Utility.isOnScreen(base.Position, 128) && Sprite.currentFrame == 0 && currentFrame == 7)
			{
				base.currentLocation.localSound("slimeHit");
			}
		}


		public override void reloadSprite(bool onlyAppearance = false)
		{
			Sprite.SpriteWidth = ModEntry.Config.PlantGirlBossWidth;
			Sprite.SpriteHeight = ModEntry.Config.PlantGirlBossHeight;
			Sprite.LoadTexture(ModEntry.GetBossTexture(GetType()));
			Sprite.UpdateSourceRect();
			Scale = ModEntry.Config.PlantGirlBossScale;
			Sprite.framesPerAnimation = 4;
			ignoreMovementAnimations = true;
			speed = 1;
			defaultAnimationInterval.Value = 150;
		}

		int skipLeft = 0;
		int skipRight = 0;
		int skipUp = 0;
		int skipDown = 0;
		bool skipX = false;
		private void moveBoss(GameTime time)
		{
			base.IsWalkingTowardPlayer = true;
			Point standingPixel = base.StandingPixel;
			Point standingPixel2 = Player.StandingPixel;
			if (currentLocation.isCollidingPosition(nextPosition(0), Game1.viewport, isFarmer: false, DamageToFarmer, isGlider.Value, this))
			{ skipUp = 1000; }
			if (currentLocation.isCollidingPosition(nextPosition(3), Game1.viewport, isFarmer: false, DamageToFarmer, isGlider.Value, this))
			{ skipLeft = 1000; }
			if (currentLocation.isCollidingPosition(nextPosition(2), Game1.viewport, isFarmer: false, DamageToFarmer, isGlider.Value, this))
			{ skipDown = 1000; }
			if (currentLocation.isCollidingPosition(nextPosition(1), Game1.viewport, isFarmer: false, DamageToFarmer, isGlider.Value, this))
			{ skipRight = 1000; }
			if (Math.Abs(StandingPixel.X - Player.StandingPixel.X) > Math.Abs(StandingPixel.Y - Player.StandingPixel.Y)) { skipX = true; } else { skipX = false; }
			if ((!skipX && (skipLeft <= 0 || skipRight <= 0)) || (Math.Abs(standingPixel2.Y - standingPixel.Y) > 192 && (skipLeft <= 0 || skipRight <= 0)) || (skipDown > 0 && skipUp > 0))
			{
				if (((standingPixel2.X - standingPixel.X > 0 && standingPixel2.X - standingPixel.X < 256)
					|| (standingPixel2.X - standingPixel.X < 0 && standingPixel2.X - standingPixel.X < -256)) && (skipLeft <= 0 || skipRight <= 0))
				{
					if (skipLeft <= 0) { SetMovingLeft(b: true); } else { SetMovingRight(b: true); }
				}
				else
				{
					if (skipRight <= 0) { SetMovingRight(b: true); } else { SetMovingRight(b: true); }
				}
			}
			else if (standingPixel2.Y - standingPixel.Y > 1.5 * height
				&& (skipUp <= 0 || skipDown <= 0))
			{
				if (skipUp <= 0) { SetMovingUp(b: true); } else { SetMovingDown(b: true); }
			}
			else
			{
				if (skipDown <= 0) SetMovingDown(b: true); else { SetMovingUp(b: true); }
			}

			if (skipRight > 0) skipRight -= time.ElapsedGameTime.Milliseconds;
			if (skipLeft > 0) skipLeft -= time.ElapsedGameTime.Milliseconds;
			if (skipUp > 0) skipUp -= time.ElapsedGameTime.Milliseconds;
			if (skipDown > 0) skipDown -= time.ElapsedGameTime.Milliseconds;
		}

		double timeShield = 10000;
		float Distance = 150;
		float AngularSpeed = 0.7f;
		float angle = 0;
		int[] isShieldExists = new int[4] { 0, 0, 0, 0 };
		BossProjectile[] projectileSh = new BossProjectile[4];
		public override void behaviorAtGameTick(GameTime time)
		{
			switch (Health)
			{
				case int n when (n >= 0.9 * MaxHealth):
					addedSpeed = 0;
					break;

				case int n when (n < 0.9 * MaxHealth && n >= 0.1 * MaxHealth):
					addedSpeed = 0.5f * difficulty;
					break;

				case int n when (n < 0.1 * MaxHealth):
					addedSpeed = 1 * difficulty;
					break;
			}
			faceGeneralDirection(base.Player.Position);
			currentLocation.characters[^1].moveTowardPlayer(moveTowardPlayerThreshold.Value);
			if (!withinPlayerThreshold(3))
			{
				speed = 1; base.IsWalkingTowardPlayer = true;
			}
			else
			{
				speed = 2; base.IsWalkingTowardPlayer = false;
			}
			if (Health <= 0)
			{
				return;
			}

			if (difficulty > 0.2)
				if (withinPlayerThreshold(moveTowardPlayerThreshold.Value))
				{
					lastIceBall = Math.Max(0f, lastIceBall - time.ElapsedGameTime.Milliseconds);

					if (lastIceBall == 0f)
					{
						Vector2 trajectory = ModEntry.VectorFromDegree(Game1.random.Next(0, 360)) * 10f;
						Vector2 startPosition = new Vector2(position.X, position.Y);
						currentLocation.projectiles.Add(new BossProjectile((int)Math.Round(10 * difficulty), 11, 0, 0, 0f, trajectory.X, trajectory.Y, startPosition, "", "", "", true, false, currentLocation, this, null, null, "13"));

						projectileCount++;

						if (projectileCount >= Math.Round((ModEntry.IsLessThanHalfHealth(this) ? 2.5 : 1.5) * Math.Round(3 * difficulty, MidpointRounding.AwayFromZero), MidpointRounding.AwayFromZero))
						{
							projectileCount = 0;
							lastIceBall = Game1.random.Next((int)Math.Round(1200 / difficulty), (int)Math.Round(3500 / difficulty));
						}
						else
						{
							lastIceBall = 100;
						}
						if (lastIceBall != 0f && Game1.random.NextDouble() < 0.05)
						{
							Halt();
							setTrajectory((int)Utility.getVelocityTowardPlayer(Utility.Vector2ToPoint(getStandingPosition()), 8f, Player).X, (int)(-(int)Utility.getVelocityTowardPlayer(Utility.Vector2ToPoint(getStandingPosition()), 8f, Player).Y));
						}
					}
				}


			if (currentIceBallShield < 5)
			{
				if (withinPlayerThreshold(6))
				{
					maxIceBallShield = (int)Math.Round(difficulty * 2, MidpointRounding.AwayFromZero);
					if (maxIceBallShield > 4) maxIceBallShield = 4;
					shieldTimer += time.ElapsedGameTime.Milliseconds;

					if (shieldTimer > 300000)
					{
						shieldTimer = 0;
						for (int i = 0; i < 4; i++)
						{
							if (projectileSh[i] is not null && isShieldExists[i] == 1)
							{
								projectileSh[i].maxTravelDistance.Value = 1; projectileSh[i].travelDistance = 3;
								Game1.createRadialDebris(currentLocation, 12, (int)(projectileSh[i].position.X + 32f) / 64, (int)(projectileSh[i].position.Y + 32f) / 64, 6, resource: false);
								Game1.Multiplayer.broadcastSprites(currentLocation, new TemporaryAnimatedSprite(362, Game1.random.Next(30, 90), 6, 1, projectileSh[i].position.Value, flicker: false, Game1.random.NextBool()));
								isShieldExists[i] = 0;
							}
						}
						currentIceBallShield = 0; timeShield = 10000;
					}

					if (currentIceBallShield < maxIceBallShield && timeShield < 0)
					{
						if (isShieldExists[currentIceBallShield] != 1)
						{
							projectileSh[currentIceBallShield] = new BossProjectile((int)(5 * difficulty), 7, 0, 0, 0, 0, 0, new Vector2(Position.X, Position.Y), "", "", "", false, false, currentLocation, this, null, null, "19", false);
							projectileSh[currentIceBallShield].localScale = 0.5f;
							projectileSh[currentIceBallShield].IgnoreLocationCollision = true;
							projectileSh[currentIceBallShield].ignoreTravelGracePeriod.Value = true;
							projectileSh[currentIceBallShield].ignoreObjectCollisions.Value = true;
							projectileSh[currentIceBallShield].ignoreTreeCollisions = true;
							currentLocation.projectiles.Add(projectileSh[currentIceBallShield]);
							timeShield = 10000;
							isShieldExists[currentIceBallShield] = 1;
							currentIceBallShield += 1;
						}
						else { currentIceBallShield += 1; }
					}
				}
				timeShield -= time.ElapsedGameTime.TotalMilliseconds;
				angle += (float)time.ElapsedGameTime.TotalSeconds * AngularSpeed;
				for (int i = 0; i < maxIceBallShield; i++)
				{
					var displacedPosition = new Vector2((float)(Distance * Math.Sin(angle + Math.PI / 2 * i) + (position.X)),
						(float)(Distance * Math.Cos(angle + Math.PI / 2 * i) + (position.Y)));
					if (projectileSh[i] is not null)
					{
						projectileSh[i].position.Set(displacedPosition);
						if (projectileSh[i].destroyMe && isShieldExists[i] == 1) { currentIceBallShield = 0; isShieldExists[i] = 0; timeShield = 10000; projectileSh[i].destroyMe = false; }
					}
				}
			}

			if (difficulty > 0.45) behaviorAtGameTick2(time);
		}


		public override Rectangle GetBoundingBox()
		{
			Vector2 vector = Position;
			return new Rectangle((int)Math.Round(vector.X + width * scale.Value / 2), (int)Math.Round(vector.Y + scale.Value * height), width, (int)Math.Round(height * scale.Value * 2 / 3));
		}


		public override void draw(SpriteBatch b)
		{
		}

		public override void drawAboveAllLayers(SpriteBatch b)
		{
			if (Sprite.SpriteWidth < 64) reloadSprite();
			b.Draw(Sprite.Texture, getLocalPosition(Game1.viewport) + new Vector2(width * 2f, (float)(height * 1.5f)), new Rectangle?(Sprite.SourceRect), Color.White, 0f, new Vector2(width /*/ 2*/, height), scale.Value * 4f, flip ? SpriteEffects.FlipHorizontally : SpriteEffects.None, Math.Max(0f, drawOnTop ? 0.991f : (StandingPixel.Y / 10000f)));
			b.Draw(Game1.shadowTexture, getLocalPosition(Game1.viewport) + new Vector2(width * 2f, height * 1.5f), new Rectangle?(Game1.shadowTexture.Bounds), Color.White, 0f, new Vector2((float)Game1.shadowTexture.Bounds.Center.X, (float)Game1.shadowTexture.Bounds.Center.Y), /*3f +*/ yOffset / 20f, SpriteEffects.None, (StandingPixel.Y - 1) / 10000f);
		}

		public override int takeDamage(int damage, int xTrajectory, int yTrajectory, bool isBomb, double addedPrecision, Farmer who)
		{
			int result = base.takeDamage(damage, 0, 0, isBomb, addedPrecision, who);
			if (Health <= 0)
			{
				ModEntry.BossDeath(currentLocation, this, difficulty);

			}
			ModEntry.MakeBossHealthBar(Health, MaxHealth);
			return result;
		}

		public override void updateMovement(GameLocation location, GameTime time)
		{
			if (base.IsWalkingTowardPlayer)
			{
				if ((moveTowardPlayerThreshold.Value == -1 || withinPlayerThreshold()) && timeBeforeAIMovementAgain <= 0f && IsMonster && !isGlider.Value)
				{
					Tile tile = location.map.RequireLayer("Back").Tiles[Player.TilePoint.X, Player.TilePoint.Y];
					if (tile == null || tile.Properties.ContainsKey("NPCBarrier"))
					{
						return;
					}

					if (skipHorizontal <= 0)
					{
						if (lastPosition.Equals(base.Position) && Game1.random.NextDouble() < 0.001)
						{
							switch (FacingDirection)
							{
								case 1:
								case 3:
									if (Game1.random.NextBool())
									{
										SetMovingOnlyUp();
									}
									else
									{
										SetMovingOnlyDown();
									}

									break;
								case 0:
								case 2:
									if (Game1.random.NextBool())
									{
										SetMovingOnlyRight();
									}
									else
									{
										SetMovingOnlyLeft();
									}

									break;
							}

							skipHorizontal = 700;
							return;
						}

						bool success = false;
						bool setMoving = false;
						bool scootSuccess = false;
						if (lastPosition.X == base.Position.X)
						{
							checkHorizontalMovement(ref success, ref setMoving, ref scootSuccess, Player, location);
							if (!success) checkVerticalMovement(ref success, ref setMoving, ref scootSuccess, Player, location);
						}
						else
						{
							checkVerticalMovement(ref success, ref setMoving, ref scootSuccess, Player, location);
							if (!success) checkHorizontalMovement(ref success, ref setMoving, ref scootSuccess, Player, location);
						}
						if (success)
						{
							skipHorizontal = 500;
						}
						if (!success)
							if (lastPosition.Y == base.Position.Y)
							{
								checkVerticalMovement(ref success, ref setMoving, ref scootSuccess, Player, location);
								if (!success) checkHorizontalMovement(ref success, ref setMoving, ref scootSuccess, Player, location);
							}
							else
							{
								checkHorizontalMovement(ref success, ref setMoving, ref scootSuccess, Player, location);
								if (!success) checkVerticalMovement(ref success, ref setMoving, ref scootSuccess, Player, location);
							}

						if (!success && !setMoving)
						{
							Halt();
						}

						if (scootSuccess)
						{
							return;
						}
					}
					else
					{
						skipHorizontal -= time.ElapsedGameTime.Milliseconds;
					}
				}
			}
			else
			{
				if ((moveTowardPlayerThreshold.Value == -1 || withinPlayerThreshold()) && timeBeforeAIMovementAgain <= 0f && IsMonster && !isGlider.Value)
				{
					Tile tile = location.map.RequireLayer("Back").Tiles[Player.TilePoint.X, Player.TilePoint.Y];
					if (tile == null || tile.Properties.ContainsKey("NPCBarrier"))
					{
						return;
					}

					if (skipHorizontal <= 0)
					{
						if (lastPosition.Equals(base.Position) && Game1.random.NextDouble() < 0.001)
						{
							switch (FacingDirection)
							{
								case 1:
								case 3:
									if (Game1.random.NextBool())
									{
										SetMovingOnlyUp();
									}
									else
									{
										SetMovingOnlyDown();
									}

									break;
								case 0:
								case 2:
									if (Game1.random.NextBool())
									{
										SetMovingOnlyRight();
									}
									else
									{
										SetMovingOnlyLeft();
									}

									break;
							}

							skipHorizontal = 700;
							return;
						}

						bool success = false;
						bool setMoving = false;
						bool scootSuccess = false;
						if (lastPosition.X == base.Position.X)
						{
							checkHorizontalMovement(ref success, ref setMoving, ref scootSuccess, Player, location, true);
							if (!success) checkVerticalMovement(ref success, ref setMoving, ref scootSuccess, Player, location, true);
						}
						else
						{
							checkVerticalMovement(ref success, ref setMoving, ref scootSuccess, Player, location, true);
							if (!success) checkHorizontalMovement(ref success, ref setMoving, ref scootSuccess, Player, location, true);
						}
						if (success)
						{
							skipHorizontal = 500;
						}
						if (!success)
							if (lastPosition.Y == base.Position.Y)
							{
								checkVerticalMovement(ref success, ref setMoving, ref scootSuccess, Player, location, true);
								if (!success) checkHorizontalMovement(ref success, ref setMoving, ref scootSuccess, Player, location, true);
							}
							else
							{
								checkHorizontalMovement(ref success, ref setMoving, ref scootSuccess, Player, location, true);
								if (!success) checkVerticalMovement(ref success, ref setMoving, ref scootSuccess, Player, location, true);
							}

						if (!success && !setMoving)
						{
							Halt();
						}

						if (scootSuccess)
						{
							return;
						}
					}
					else
					{
						skipHorizontal -= time.ElapsedGameTime.Milliseconds;
					}
				}
			}

			MovePosition(time, Game1.viewport, location);
			if (base.Position.Equals(lastPosition) && base.IsWalkingTowardPlayer && withinPlayerThreshold())
			{
				noMovementProgressNearPlayerBehavior();
			}
		}

		private void checkHorizontalMovement(ref bool success, ref bool setMoving, ref bool scootSuccess, Farmer who, GameLocation location, bool inverse = false)
		{
			if (who.Position.X > base.Position.X + 16f)
			{
				if (ShouldActuallyMoveAwayFromPlayer())
				{
					if (!inverse) SetMovingOnlyLeft(); else SetMovingOnlyRight();
				}
				else
				{
					if (!inverse) SetMovingOnlyRight(); else SetMovingOnlyLeft();
				}

				setMoving = true;
				if (!location.isCollidingPosition(nextPosition(1), Game1.viewport, isFarmer: false, DamageToFarmer, isGlider.Value, this))
				{
					success = true;
				}
				else
				{
					MovePosition(Game1.currentGameTime, Game1.viewport, location);
					if (!base.Position.Equals(lastPosition))
					{
						scootSuccess = true;
					}
				}
			}

			if (success || !(who.Position.X < base.Position.X - 16f))
			{
				return;
			}

			if (ShouldActuallyMoveAwayFromPlayer())
			{
				if (!inverse) SetMovingOnlyRight(); else SetMovingOnlyLeft();
			}
			else
			{
				if (!inverse) SetMovingOnlyLeft(); else SetMovingOnlyRight();
			}

			setMoving = true;
			if (!location.isCollidingPosition(nextPosition(3), Game1.viewport, isFarmer: false, DamageToFarmer, isGlider.Value, this))
			{
				success = true;
				return;
			}

			MovePosition(Game1.currentGameTime, Game1.viewport, location);
			if (!base.Position.Equals(lastPosition))
			{
				scootSuccess = true;
			}
		}

		private void checkVerticalMovement(ref bool success, ref bool setMoving, ref bool scootSuccess, Farmer who, GameLocation location, bool inverse = false)
		{
			if (!success && who.Position.Y < base.Position.Y - 16f)
			{
				if (ShouldActuallyMoveAwayFromPlayer())
				{
					if (!inverse) SetMovingOnlyDown(); else SetMovingOnlyUp();
				}
				else
				{
					if (!inverse) SetMovingOnlyUp(); else SetMovingOnlyDown();
				}

				setMoving = true;
				if (!location.isCollidingPosition(nextPosition(0), Game1.viewport, isFarmer: false, DamageToFarmer, isGlider.Value, this))
				{
					success = true;
				}
				else
				{
					MovePosition(Game1.currentGameTime, Game1.viewport, location);
					if (!base.Position.Equals(lastPosition))
					{
						scootSuccess = true;
					}
				}
			}

			if (success || !(who.Position.Y > base.Position.Y + 16f))
			{
				return;
			}

			if (ShouldActuallyMoveAwayFromPlayer())
			{
				if (!inverse) SetMovingOnlyUp(); else SetMovingOnlyDown();
			}
			else
			{
				if (!inverse) SetMovingOnlyDown(); else SetMovingOnlyUp();
			}

			setMoving = true;
			if (!location.isCollidingPosition(nextPosition(2), Game1.viewport, isFarmer: false, DamageToFarmer, isGlider.Value, this))
			{
				success = true;
				return;
			}

			MovePosition(Game1.currentGameTime, Game1.viewport, location);
			if (!base.Position.Equals(lastPosition))
			{
				scootSuccess = true;
			}
		}

		private int slideAnimationTimer;
		public override void MovePosition(GameTime time, xTile.Dimensions.Rectangle viewport, GameLocation currentLocation)
		{
			if ((int)stunTime.Value > 0)
			{
				return;
			}

			lastPosition = base.Position;
			if (xVelocity != 0f || yVelocity != 0f)
			{
				if (double.IsNaN(xVelocity) || double.IsNaN(yVelocity))
				{
					xVelocity = 0f;
					yVelocity = 0f;
				}

				Microsoft.Xna.Framework.Rectangle boundingBox = GetBoundingBox();
				int x = boundingBox.X;
				int y = boundingBox.Y;
				int num = boundingBox.X; 
				int num2 = boundingBox.Y;
				int num3 = 1;
				bool flag = false;
				bool flag2 = false;
				if (this is SquidKid)
				{
					flag2 = true;
				}

				if (!isGlider.Value || flag2)
				{
					if (boundingBox.Width > 0 && Math.Abs((int)xVelocity) > boundingBox.Width)
					{
						num3 = (int)Math.Max(num3, Math.Ceiling((float)Math.Abs((int)xVelocity) / (float)boundingBox.Width));
					}

					if (boundingBox.Height > 0 && Math.Abs((int)yVelocity) > boundingBox.Height)
					{
						num3 = (int)Math.Max(num3, Math.Ceiling((float)Math.Abs((int)yVelocity) / (float)boundingBox.Height));
					}
				}


				for (int i = 1; i <= num3; i++)
				{
					boundingBox.X = (int)Utility.Lerp(x, num, (float)i / (float)num3);
					boundingBox.Y = (int)Utility.Lerp(y, num2, (float)i / (float)num3);
					bool glider = isGlider.Value;
					if (flag2)
					{
						glider = false;
					}

					if (currentLocation != null && currentLocation.isCollidingPosition(boundingBox, viewport, isFarmer: false, DamageToFarmer, glider, this))
					{
						flag = true;
						break;
					}
				}

				if (!flag)
				{
					base.position.X += xVelocity;
					base.position.Y -= yVelocity;
					if (Slipperiness < 1000)
					{
						xVelocity -= xVelocity / (float)Slipperiness;
						yVelocity -= yVelocity / (float)Slipperiness;
						if (Math.Abs(xVelocity) <= 0.05f)
						{
							xVelocity = 0f;
						}

						if (Math.Abs(yVelocity) <= 0.05f)
						{
							yVelocity = 0f;
						}
					}

					if (!isGlider.Value && invincibleCountdown > 0)
					{
						slideAnimationTimer -= time.ElapsedGameTime.Milliseconds;
						if (slideAnimationTimer < 0 && (Math.Abs(xVelocity) >= 3f || Math.Abs(yVelocity) >= 3f))
						{
							slideAnimationTimer = 100 - (int)(Math.Abs(xVelocity) * 2f + Math.Abs(yVelocity) * 2f);
							Game1.Multiplayer.broadcastSprites(currentLocation, new TemporaryAnimatedSprite(6, getStandingPosition() + new Vector2(-32f, -32f), Color.White * 0.75f, 8, Game1.random.NextBool(), 20f)
							{
								scale = 0.75f
							});
						}
					}
				}
				else if ((bool)isGlider.Value || Slipperiness >= 8)
				{
					if ((bool)isGlider.Value)
					{
						bool[] array = Utility.horizontalOrVerticalCollisionDirections(boundingBox, this);
						if (array[0])
						{
							xVelocity = 0f - xVelocity;
							base.position.X += Math.Sign(xVelocity);
							rotation += (float)(Math.PI + (double)Game1.random.Next(-10, 11) * Math.PI / 500.0);
						}

						if (array[1])
						{
							yVelocity = 0f - yVelocity;
							base.position.Y -= Math.Sign(yVelocity);
							rotation += (float)(Math.PI + (double)Game1.random.Next(-10, 11) * Math.PI / 500.0);
						}
					}

					if (Slipperiness < 1000)
					{
						xVelocity -= xVelocity / (float)Slipperiness / 4f;
						yVelocity -= yVelocity / (float)Slipperiness / 4f;
						if (Math.Abs(xVelocity) <= 0.05f)
						{
							xVelocity = 0f;
						}

						if (Math.Abs(yVelocity) <= 0.051f)
						{
							yVelocity = 0f;
						}
					}
				}
				else
				{
					xVelocity -= xVelocity / (float)Slipperiness;
					yVelocity -= yVelocity / (float)Slipperiness;
					if (Math.Abs(xVelocity) <= 0.05f)
					{
						xVelocity = 0f;
					}

					if (Math.Abs(yVelocity) <= 0.05f)
					{
						yVelocity = 0f;
					}
				}

				if ((bool)isGlider.Value)
				{
					return;
				}
			}

			if (moveUp)
			{
				if (((!Game1.eventUp || Game1.IsMultiplayer) && !currentLocation.isCollidingPosition(nextPosition(0), viewport, isFarmer: false, DamageToFarmer, isGlider.Value, this)) || isCharging)
				{
					base.position.Y -= (float)base.speed + addedSpeed;
					if (!ignoreMovementAnimations)
					{
						Sprite.AnimateUp(time);
					}
				}
				else
				{
					Microsoft.Xna.Framework.Rectangle position = nextPosition(0);
					position.Width /= 4;
					bool flag3 = currentLocation.isCollidingPosition(position, viewport, isFarmer: false, DamageToFarmer, isGlider.Value, this);
					position.X += position.Width * 3;
					bool flag4 = currentLocation.isCollidingPosition(position, viewport, isFarmer: false, DamageToFarmer, isGlider.Value, this);
					if (flag3 && !flag4 && !currentLocation.isCollidingPosition(nextPosition(1), viewport, isFarmer: false, DamageToFarmer, isGlider.Value, this))
					{
						base.position.X += (float)base.speed * ((float)time.ElapsedGameTime.Milliseconds / 64f);
					}
					else if (flag4 && !flag3 && !currentLocation.isCollidingPosition(nextPosition(3), viewport, isFarmer: false, DamageToFarmer, isGlider.Value, this))
					{
						base.position.X -= (float)base.speed * ((float)time.ElapsedGameTime.Milliseconds / 64f);
					}

					if (!currentLocation.isTilePassable(nextPosition(0), viewport) || !base.willDestroyObjectsUnderfoot)
					{
						Halt();
					}
					else if (base.willDestroyObjectsUnderfoot)
					{
						if (currentLocation.characterDestroyObjectWithinRectangle(nextPosition(0), showDestroyedObject: true))
						{
							currentLocation.playSound("stoneCrack");
							base.position.Y -= (float)base.speed + addedSpeed;
						}
						else
						{
							blockedInterval += time.ElapsedGameTime.Milliseconds;
						}
					}

					onCollision?.Invoke(currentLocation);
				}
			}
			else if (moveRight)
			{
				if (((!Game1.eventUp || Game1.IsMultiplayer) && !currentLocation.isCollidingPosition(nextPosition(1), viewport, isFarmer: false, DamageToFarmer, isGlider.Value, this)) || isCharging)
				{
					base.position.X += (float)base.speed + addedSpeed;
					if (!ignoreMovementAnimations)
					{
						Sprite.AnimateRight(time);
					}
				}
				else
				{
					Microsoft.Xna.Framework.Rectangle position2 = nextPosition(1);
					position2.Height /= 4;
					bool flag5 = currentLocation.isCollidingPosition(position2, viewport, isFarmer: false, DamageToFarmer, isGlider.Value, this);
					position2.Y += position2.Height * 3;
					bool flag6 = currentLocation.isCollidingPosition(position2, viewport, isFarmer: false, DamageToFarmer, isGlider.Value, this);
					if (flag5 && !flag6 && !currentLocation.isCollidingPosition(nextPosition(2), viewport, isFarmer: false, DamageToFarmer, isGlider.Value, this))
					{
						base.position.Y += (float)base.speed * ((float)time.ElapsedGameTime.Milliseconds / 64f);
					}
					else if (flag6 && !flag5 && !currentLocation.isCollidingPosition(nextPosition(0), viewport, isFarmer: false, DamageToFarmer, isGlider.Value, this))
					{
						base.position.Y -= (float)base.speed * ((float)time.ElapsedGameTime.Milliseconds / 64f);
					}

					if (!currentLocation.isTilePassable(nextPosition(1), viewport) || !base.willDestroyObjectsUnderfoot)
					{
						Halt();
					}
					else if (base.willDestroyObjectsUnderfoot)
					{
						if (currentLocation.characterDestroyObjectWithinRectangle(nextPosition(1), showDestroyedObject: true))
						{
							currentLocation.playSound("stoneCrack");
							base.position.X += (float)base.speed + addedSpeed;
						}
						else
						{
							blockedInterval += time.ElapsedGameTime.Milliseconds;
						}
					}

					onCollision?.Invoke(currentLocation);
				}
			}
			else if (moveDown)
			{
				if (((!Game1.eventUp || Game1.IsMultiplayer) && !currentLocation.isCollidingPosition(nextPosition(2), viewport, isFarmer: false, DamageToFarmer, isGlider.Value, this)) || isCharging)
				{
					base.position.Y += (float)base.speed + addedSpeed;
					if (!ignoreMovementAnimations)
					{
						Sprite.AnimateDown(time);
					}
				}
				else
				{
					Microsoft.Xna.Framework.Rectangle position3 = nextPosition(2);
					position3.Width /= 4;
					bool flag7 = currentLocation.isCollidingPosition(position3, viewport, isFarmer: false, DamageToFarmer, isGlider.Value, this);
					position3.X += position3.Width * 3;
					bool flag8 = currentLocation.isCollidingPosition(position3, viewport, isFarmer: false, DamageToFarmer, isGlider.Value, this);
					if (flag7 && !flag8 && !currentLocation.isCollidingPosition(nextPosition(1), viewport, isFarmer: false, DamageToFarmer, isGlider.Value, this))
					{
						base.position.X += (float)base.speed * ((float)time.ElapsedGameTime.Milliseconds / 64f);
					}
					else if (flag8 && !flag7 && !currentLocation.isCollidingPosition(nextPosition(3), viewport, isFarmer: false, DamageToFarmer, isGlider.Value, this))
					{
						base.position.X -= (float)base.speed * ((float)time.ElapsedGameTime.Milliseconds / 64f);
					}

					if (!currentLocation.isTilePassable(nextPosition(2), viewport) || !base.willDestroyObjectsUnderfoot)
					{
						Halt();
					}
					else if (base.willDestroyObjectsUnderfoot)
					{
						if (currentLocation.characterDestroyObjectWithinRectangle(nextPosition(2), showDestroyedObject: true))
						{
							currentLocation.playSound("stoneCrack");
							base.position.Y += (float)base.speed + addedSpeed;
						}
						else
						{
							blockedInterval += time.ElapsedGameTime.Milliseconds;
						}
					}

					onCollision?.Invoke(currentLocation);
				}
			}
			else if (moveLeft)
			{
				if (((!Game1.eventUp || Game1.IsMultiplayer) && !currentLocation.isCollidingPosition(nextPosition(3), viewport, isFarmer: false, DamageToFarmer, isGlider.Value, this)) || isCharging)
				{
					base.position.X -= (float)base.speed + addedSpeed;
					if (!ignoreMovementAnimations)
					{
						Sprite.AnimateLeft(time);
					}
				}
				else
				{
					Microsoft.Xna.Framework.Rectangle position4 = nextPosition(3);
					position4.Height /= 4;
					bool flag9 = currentLocation.isCollidingPosition(position4, viewport, isFarmer: false, DamageToFarmer, isGlider.Value, this);
					position4.Y += position4.Height * 3;
					bool flag10 = currentLocation.isCollidingPosition(position4, viewport, isFarmer: false, DamageToFarmer, isGlider.Value, this);
					if (flag9 && !flag10 && !currentLocation.isCollidingPosition(nextPosition(2), viewport, isFarmer: false, DamageToFarmer, isGlider.Value, this))
					{
						base.position.Y += (float)base.speed * ((float)time.ElapsedGameTime.Milliseconds / 64f);
					}
					else if (flag10 && !flag9 && !currentLocation.isCollidingPosition(nextPosition(0), viewport, isFarmer: false, DamageToFarmer, isGlider.Value, this))
					{
						base.position.Y -= (float)base.speed * ((float)time.ElapsedGameTime.Milliseconds / 64f);
					}

					if (!currentLocation.isTilePassable(nextPosition(3), viewport) || !base.willDestroyObjectsUnderfoot)
					{
						Halt();
					}
					else if (base.willDestroyObjectsUnderfoot)
					{
						if (currentLocation.characterDestroyObjectWithinRectangle(nextPosition(3), showDestroyedObject: true))
						{
							currentLocation.playSound("stoneCrack");
							base.position.X -= (float)base.speed + addedSpeed;
						}
						else
						{
							blockedInterval += time.ElapsedGameTime.Milliseconds;
						}
					}

					onCollision?.Invoke(currentLocation);
				}
			}
			else if (!ignoreMovementAnimations)
			{
				if (moveUp)
				{
					Sprite.AnimateUp(time);
				}
				else if (moveRight)
				{
					Sprite.AnimateRight(time);
				}
				else if (moveDown)
				{
					Sprite.AnimateDown(time);
				}
				else if (moveLeft)
				{
					Sprite.AnimateLeft(time);
				}
			}

			if (blockedInterval >= 5000)
			{
				base.speed = 4;
				isCharging = true;
				blockedInterval = 0;
			}

			if (DamageToFarmer <= 0 || !(Game1.random.NextDouble() < 0.00033333333333333332))
			{
				return;
			}

			string name = base.Name;
			if (!(name == "Shadow Guy"))
			{
				if (name == "Ghost")
				{
					currentLocation.playSound("ghost");
				}
			}
			else if (Game1.random.NextDouble() < 0.3)
			{
				if (Game1.random.NextBool())
				{
					currentLocation.playSound("grunt");
				}
				else
				{
					currentLocation.playSound("shadowpeep");
				}
			}
		}

		public override void updateFaceTowardsFarmer(GameTime time, GameLocation location) { }

		public override void Halt()
		{
		}

		public void createTentacle(GameLocation location, int i, Vector2 shot_velocity, NetPosition position, int width, int height, float difficulty, bool isPull = true)
		{
			bool one = i < 4;
			bool two = i % 4 < 2;
			bool three = i % 2 == 0;

			Vector2 v = new((three ? shot_velocity.X : shot_velocity.Y) * (one ? -1 : 1), (three ? shot_velocity.Y : shot_velocity.X) * (two ? -1 : 1));
			Vector2 startPosition = new Vector2(position.X + 0.1f * width, position.Y + 0.5f * height);
			if (v.Y < position.Y + +0.5f * height)
			{
				BasicProjectile projectile = new BossProjectile((int)(5 * difficulty), 7/*766*/, 0, 1, 0.196349546f, v.X, v.Y, startPosition, "", "", "", false, false, currentLocation, this, null, null/*766*/, "18", isPull)
				{
					IgnoreLocationCollision = true,
					localScale = 0.2f
				};
				projectile.ignoreTravelGracePeriod.Value = true;
				projectile.maxTravelDistance.Value = (int)Math.Round(412 * difficulty * 2 / 3);
				location.projectiles.Add(projectile);
			}
		}

		public void behaviorAtGameTick2(GameTime time)
		{
			if (Health <= 0)
			{
				return;
			}

			timeUntilNextAttack -= time.ElapsedGameTime.Milliseconds;

			FarmerCollection.Enumerator enumerator = currentLocation.farmers.GetEnumerator();
			while (enumerator.MoveNext())
			{
				if (enumerator.Current.currentLocation == currentLocation && enumerator.Current.GetBoundingBox().Intersects(GetBoundingBox()))
				{
					enumerator.Current.takeDamage((int)Math.Round(12 * difficulty), true, null);
					totalFireTime = 0;
					nextFireTime = 10;
					attackState.Set(0);
					timeUntilNextAttack = Game1.random.Next(1000, 2000);
				}
			}

			if (attackState.Value == 0 && withinPlayerThreshold(20))
			{
				firing.Set(false);
				if (timeUntilNextAttack < 0)
				{
					timeUntilNextAttack = 0;
					attackState.Set(1);
					nextFireTime = 1;
					totalFireTime = 3000;
					return;
				}
			}
			else if (totalFireTime > 0)
			{
				Farmer player = Player;
				if (!firing.Value)
				{
					if (player != null)
					{		
					}
				}
				totalFireTime -= time.ElapsedGameTime.Milliseconds;
				if (nextFireTime > 0)
				{
					nextFireTime -= time.ElapsedGameTime.Milliseconds;
					if (nextFireTime <= 0)
					{
						if (!firing.Value)
						{
							firing.Set(true);
						}
						float fire_angle = 0f;
						switch (facingDirection.Value)
						{
							case 0:
								fire_angle = 180f;
								break;
							case 1:
								fire_angle = 0f;
								break;
							case 2:
								fire_angle = 270f;
								break;
							case 3:
								fire_angle = 0f;
								break;
						}
						fire_angle += (float)Math.Sin((double)(totalFireTime / 1000f * 360f) * Math.PI / 360.0) * 25f;//180 to 360
						Vector2 shot_velocity = new((float)Math.Cos((double)fire_angle * Math.PI / 180.0), -(float)Math.Sin((double)fire_angle * Math.PI / 180.0));

						shot_velocity *= 5f;

						for (int i = 0; i < 8; i++)
						{
							if (facingDirection.Value == 2 || (Sprite.currentFrame >= 0 && Sprite.currentFrame < 4))
							{
								if (i != 4 && i != 5 && i != 0 && i != 1)
								{
									shot_velocity = new((float)Math.Cos((double)fire_angle * Math.PI / 180.0) + 1, -(float)Math.Sin((double)fire_angle * Math.PI / 180.0) + 1);
									if (difficulty < 1.6) createTentacle(currentLocation, i, shot_velocity, position, width, height, difficulty, false);
									else
										createTentacle(currentLocation, i, shot_velocity, position, width, height, difficulty);
									i++;
									if (ModEntry.IsLessThanHalfHealth(this) && difficulty > 1.9)
									{
										i--;
									}
								}
							}
							else
							{
								if (difficulty < 1.6) createTentacle(currentLocation, i, shot_velocity, position, width, height, difficulty, false);
								else
									createTentacle(currentLocation, i, shot_velocity, position, width, height, difficulty);
								i++;
								if (ModEntry.IsLessThanHalfHealth(this) && difficulty > 1.9)
								{
									if (i != 6 && i != 7 && i != 2 && i != 3)
										i--;
								}
							}

							nextFireTime = 10;
						}
					}
					if (totalFireTime <= 0)
					{
						totalFireTime = 0;
						nextFireTime = 5;
						attackState.Set(0);
						timeUntilNextAttack = 0;
					}
				}
			}
		}
	}
}
